"""URN module for aiowebdav2."""

from re import sub
from urllib.parse import quote, unquote, urlsplit


class Urn:
    """Class for URN representation."""

    separate = "/"

    def __init__(self, path: str, *, directory: bool = False) -> None:
        """Class for URN representation."""
        self._path = quote(path)
        expressions = r"/\.+/", "/+"
        for expression in expressions:
            self._path = sub(expression, Urn.separate, self._path)

        if not self._path.startswith(Urn.separate):
            self._path = f"{Urn.separate}{self._path}"

        if directory and not self._path.endswith(Urn.separate):
            self._path = f"{self._path}{Urn.separate}"

    def __str__(self) -> str:
        """Return string representation of URN."""
        return self.path()

    def path(self) -> str:
        """Return path."""
        return unquote(self._path)

    def quote(self) -> str:
        """Return quoted path."""
        return self._path

    def filename(self) -> str:
        """Return filename."""
        path_split = self._path.split(Urn.separate)
        name = path_split[-2] + Urn.separate if path_split[-1] == "" else path_split[-1]
        return unquote(name)

    def parent(self) -> str:
        """Return parent path."""
        path_split = self._path.split(Urn.separate)
        nesting_level = self.nesting_level()
        parent_path_split = path_split[:nesting_level]
        parent = (
            self.separate.join(parent_path_split)
            if nesting_level != 1
            else Urn.separate
        )
        if not parent.endswith(Urn.separate):
            return unquote(parent + Urn.separate)
        return unquote(parent)

    def nesting_level(self) -> int:
        """Return nesting level."""
        return self._path.count(Urn.separate, 0, -1)

    def is_dir(self) -> bool:
        """Return True if URN is directory."""
        return self._path[-1] == Urn.separate

    @staticmethod
    def normalize_path(path: str) -> str:
        """Normalize path."""
        result = sub("/{2,}", "/", path)
        return result if len(result) < 1 or result[-1] != Urn.separate else result[:-1]

    @staticmethod
    def compare_path(path_a: str, href: str) -> bool:
        """Compare paths."""
        unquoted_path = Urn.separate + unquote(urlsplit(href).path)
        return Urn.normalize_path(path_a) == Urn.normalize_path(unquoted_path)



"""Exceptions for aiowebdav2."""


class WebDavError(Exception):
    """Base class for all webdav exceptions."""


class NotValidError(WebDavError):
    """Base class for all not valid exceptions."""


class OptionNotValidError(NotValidError):
    """Exception for not valid options."""

    def __init__(self, name: str, value: str, ns: str = "") -> None:
        """Exception for not valid options."""
        self.name = name
        self.value = value
        self.ns = ns

    def __str__(self) -> str:
        """Return string representation of exception."""
        return f"Option ({self.ns}{self.name}={self.value}) have invalid name or value"


class CertificateNotValidError(NotValidError):
    """Exception for not valid certificate."""


class NotFoundError(WebDavError):
    """Base class for all not found exceptions."""


class LocalResourceNotFoundError(NotFoundError):
    """Exception for not found local resource."""

    def __init__(self, path: str) -> None:
        """Exception for not found local resource."""
        self.path = path

    def __str__(self) -> str:
        """Return string representation of exception."""
        return f"Local file: {self.path} not found"


class RemoteResourceNotFoundError(NotFoundError):
    """Exception for not found remote resource."""

    def __init__(self, path: str) -> None:
        """Exception for not found remote resource."""
        self.path = path

    def __str__(self) -> str:
        """Return string representation of exception."""
        return f"Remote resource: {self.path} not found"


class RemoteParentNotFoundError(NotFoundError):
    """Exception for not found remote parent."""

    def __init__(self, path: str) -> None:
        """Exception for not found remote parent."""
        self.path = path

    def __str__(self) -> str:
        """Return string representation of exception."""
        return f"Remote parent for: {self.path} not found"


class MethodNotSupportedError(WebDavError):
    """Exception for not supported method."""

    def __init__(self, name: str, server: str) -> None:
        """Exception for not supported method."""
        self.name = name
        self.server = server

    def __str__(self) -> str:
        """Return string representation of exception."""
        return f"Method '{self.name}' not supported for {self.server}"


class ConnectionExceptionError(WebDavError):
    """Exception for connection error."""

    def __init__(self, exception: Exception) -> None:
        """Exception for connection error."""
        self.exception = exception

    def __str__(self) -> str:
        """Return string representation of exception."""
        return self.exception.__str__()


class NoConnectionError(WebDavError):
    """Exception for no connection."""

    def __init__(self, hostname: str) -> None:
        """Exception for no connection."""
        self.hostname = hostname

    def __str__(self) -> str:
        """Return string representation of exception."""
        return f"No connection with {self.hostname}"


# This exception left only for supporting original library interface.
class NotConnectionError(WebDavError):
    """Exception for no connection."""

    def __init__(self, hostname: str) -> None:
        """Exception for no connection."""
        self.hostname = hostname

    def __str__(self) -> str:
        """Return string representation of exception."""
        return f"No connection with {self.hostname}"


class ResponseErrorCodeError(WebDavError):
    """Exception for response error code."""

    def __init__(self, url: str, code: int, message: str) -> None:
        """Exception for response error code."""
        self.url = url
        self.code = code
        self.message = message

    def __str__(self) -> str:
        """Return string representation of exception."""
        return (
            f"Request to {self.url} failed with code {self.code} "
            f"and message: {self.message}"
        )


class NotEnoughSpaceError(WebDavError):
    """Exception for not enough space on the server."""

    def __init__(self) -> None:
        """Exception for not enough space on the server."""
        self.message = "Not enough space on the server"

    def __str__(self) -> str:
        """Return string representation of exception."""
        return self.message


class ResourceLockedError(WebDavError):
    """Exception for locked resource."""

    def __init__(self, path: str) -> None:
        """Exception for locked resource."""
        self.path = path

    def __str__(self) -> str:
        """Return string representation of exception."""
        return f"Resource {self.path} locked"


class UnauthorizedError(WebDavError):
    """Exception for unauthorized user."""

    def __init__(self, path: str) -> None:
        """Exception for unauthorized user."""
        self.path = path

    def __str__(self) -> str:
        """Return string representation of exception."""
        return f"Unauthorized access to {self.path}"
